@@ -33,5 +33,5 @@ class CompiledPrefabError(PrefabError):
     pass
 
 
-class PrefabTypeError(PrefabError, TypeError):
+class FrozenPrefabError(PrefabError):
     pass
